///*****************************************************************************************
///Virtual Boy (VB) USB Link cable Arduino code
///Kevin Mellott. 2017.  http://www.mellottsvrpage.com
///
///This code is intended for my custom USB to link cable adapter which was initially based
///on work I did for the Nintendo Power Glove many years ago.  See website for more details.
///SPI code originally by Nick Gammon, Feb. 2011.  http://www.gammon.com.au/spi
///
///*****************************************************************************************
#include <TimerOne.h>
#include <SPI.h>
#include <util/crc16.h>

//#define DEBUG     1   //If defined serial debug messages print to console
//Pin mappings for Arduino Nano
#define MS_SEL    16  //Flips VB Data_In Data_out on SPI MOSI, MISO lines so we can use HW SPI
#define redLED    14  //RED
#define blueLED   15  //BLUE
#define greenLED  3   //GREEN
#define amberLED  4   //AMBER
#define VBSyncClk 9   //Clock from MCU to VB for syncing displays 
#define BUFMAX    5  //small serial and VB buffers
#define CTRL      8   //Control line
#define VBSyncIn  7   //Sync from VB

#define MST_1     1
#define SLV_2     2
#define MST_3     3   //service as master based on VB control line being low

//Globals.  Bad programmer. It's just easier this way.
char* ver="VB_USB_LC 0.2\0";
byte serialIncomingByte=0;
byte mode3Serial=0;
byte vbReturnByte=0;
char serialInputBuf [BUFMAX];
char cmdIndex=0;
volatile byte pos=0;
volatile boolean spi_slave_process_it=false;
volatile boolean masterMode=true;
volatile char mode=0; //0-master normal, 1-slave normal, 3-master service on control low
volatile boolean ctrlBit=true, myCtrlBit=true;
volatile byte cmdByte=0, myCmdByte=0;
volatile byte checkSum=0, myCheckSum=0;
//Message packet variables
const byte NUMBYTES = 4;
byte receivedBytes[NUMBYTES];
byte numReceived = 0;
boolean newData = false;


//Recongized Commands
#define MST    (serialInputBuf[0]=='M' && serialInputBuf[1]=='S' && serialInputBuf[2]=='T')
#define SLV    (serialInputBuf[0]=='S' && serialInputBuf[1]=='L' && serialInputBuf[2]=='V')
#define VER    (serialInputBuf[0]=='V' && serialInputBuf[1]=='E' && serialInputBuf[2]=='R')

void setup (void)
{
  //Setup serial port
  Serial.begin (115200,SERIAL_8N1); 

  #ifdef DEBUG
    Serial.println ("VB USB Link cable lives!");
  #endif

  //Setup LED pins
  pinMode(redLED, OUTPUT);
  digitalWrite(redLED,LOW);
  pinMode(greenLED, OUTPUT);
  digitalWrite(greenLED,LOW);
  pinMode(blueLED, OUTPUT);
  digitalWrite(blueLED,LOW);
  pinMode(amberLED, OUTPUT);
  digitalWrite(amberLED,LOW);

  //Setup control pin as input
  //digitalWrite(CTRL,LOW);
  pinMode(CTRL,INPUT_PULLUP); //Change pin mode to emulate open drain or open collector output

  //Setup Timer output for VB sync at 50Hz.  MCU always drives VB through VB Sync_In pin.  
  pinMode(VBSyncClk, OUTPUT);  
  digitalWrite(VBSyncClk,LOW);  
  Timer1.initialize(10000); //50Hz Sync clock for displays. Triggers at 100Hz for toggle
  Timer1.attachInterrupt( timerIsr ); // attach the service routine here

  //Setup SPI switch direction (default low -> NC->COM).  This is flipped depending
  //on VB master or slave modes.  
  pinMode(MS_SEL, OUTPUT);
  digitalWrite(MS_SEL,LOW);

  //default to SPI master at 125kHz SPI clock rate to VB. 50kHz from VB to MCU when
  //MCU is slave and VB Master.  
  if(masterMode){
    mode=1;
    digitalWrite(MS_SEL,LOW);
    setMasterLED();
    setupMasterSPI();
  }else{
    mode=2;
    digitalWrite(MS_SEL,HIGH);
    setSlaveLED();
    setupSlaveSPI();
  }
    
}

// main loop
void loop (void)
{
  //Interrupt processing for SPI Slave receive.  If we're inside this if statement then VB 
  //sent a byte to us as a Master device and MCU is in slave mode.  
  if (spi_slave_process_it)
  {
    spi_slave_process_it = false; 
    myCmdByte=checkControl(ctrlBit);
    byte buf[]={0x3C,myCmdByte,vbReturnByte,0x3E}; //vbReturnByte was set in the interrupt
    if(vbReturnByte!=0x00){Serial.write(buf,NUMBYTES);  setGreenLED();}//sends out serial to other VB via USB link cable.  
    //At this point our previous SPDR was shifted out already so lets set to 0
    SPDR=0x00;
  }  // end of flag set

  recvBytesWithStartEndMarkers();

  //If we have newData then there's a possible command to process
  if(newData==true){
    #ifdef DEBUG
      Serial.println("New Data");
    #endif
    //Parse data
    cmdByte=(receivedBytes[0] & 0x0F);    //commands are lowest 4 bits
    ctrlBit=(receivedBytes[0] & 0x80)>>7;  //control bit is highest bit
    serialIncomingByte=receivedBytes[1];
    #ifdef DEBUG
      Serial.print("cmdByte: ");
      Serial.println(cmdByte,HEX);
      Serial.print("ctrlBit: ");
      Serial.println(ctrlBit,HEX);
      Serial.print("dataByte: ");
      Serial.println(serialIncomingByte,HEX);
    #endif

    //Process cmds.  1 is Master mode. 2 is Slave mode.  4 is print version.
    switch(cmdByte)
    {
      case 1:
        mode=1;
        masterMode=true;
        //Set SPI/MCU into master mode.  VB should be slave
        digitalWrite(MS_SEL,LOW);
        setMasterLED();
        setupMasterSPI();
      break;

      case 2:
        mode=2;
        masterMode=false;
        //Set SPI/MCU into slave mode.  VB should be master
        digitalWrite(MS_SEL,HIGH);
        setSlaveLED();
        setupSlaveSPI();
      break;

      case 3:
        //Mode 3 is Cable as Master which reacts to low pulses from control VB pin
        mode=3;
        masterMode=true;
        //Set SPI/MCU into master mode.  VB should be slave
        digitalWrite(MS_SEL,LOW);
        setMasterLED();
        setupMasterSPI();
      break;

      case 4:
        Serial.println(ver);
      break;

      default:
      break;  
    }

    //Drive Control line to our VB if need be. Also read it to send back to other VB.
    myCmdByte=checkControl(ctrlBit);
  
    //If we're in Master mode then we need to transmit the data byte to our VB and send any return bytes out the serial port
    //to the other VB.
    if(masterMode==true && mode==1){
      SPI.beginTransaction (SPISettings (125000, MSBFIRST, SPI_MODE3));//MODE 3 matches VB output when it's Master
      vbReturnByte=SPI.transfer(serialIncomingByte); //<cmd/ctrl data> with no commas. 4 bytes with header and tail. Blocking transfer.
      SPI.endTransaction();  
      byte buf[]={0x3C,myCmdByte,vbReturnByte,0x3E};
      if(vbReturnByte!=0x00){Serial.write(buf,NUMBYTES); setGreenLED();}
    }else if(masterMode==false && mode==2){
      //else we're in slave mode and need to stuff SPDR which will be pushed out when the master writes something to us.
      #ifdef DEBUG
        Serial.print("SPDR Out: ");
        Serial.println(serialIncomingByte,HEX);
      #endif
      if(serialIncomingByte!=0x00){
        SPDR=serialIncomingByte; 
       }
    }else if(masterMode==true && mode==3){
      //Do nothing as serialIncomingByte should have any data that may be ready to go out
      mode3Serial=serialIncomingByte;
    }
    cmdByte=0;
    newData=false;
    myCmdByte=0;
    vbReturnByte=0;
    serialIncomingByte=0;
  }
  
  if(mode==3 && digitalRead(CTRL)==0){
     #ifdef DEBUG
       Serial.print("mode3Serial: ");
       Serial.println(mode3Serial,HEX);
     #endif
     //Software polling of CTRL pin to service VB.  Here the USB link cable is waiting as a Master for the VB to tell it
     //that data is ready to be serviced.  If we end up in this case then the VB/game is in slave mode and tells
     //the USB linkcable that it has data ready by pulsing CTRL low.  Then the cable responds and reads the data.  
     //That data is then sent out the serial port to eventually end up at another VB/USB linkcable somewhere out in the world.  
     SPI.beginTransaction (SPISettings (125000, MSBFIRST, SPI_MODE3));//MODE 3 matches VB output when it's Master
     vbReturnByte=SPI.transfer(mode3Serial); //<cmd/ctrl data> with no commas. 4 bytes with header and tail. Blocking transfer.
     SPI.endTransaction();  
     myCmdByte=0x00; //we're here because CTRL is low and we're not using anything else in the byte yet
     byte buf[]={0x3C,myCmdByte,vbReturnByte,0x3E};
     if(vbReturnByte!=0x00){Serial.write(buf,NUMBYTES); setGreenLED();}
     //mode3Serial=0; 
     //vbReturnByte=0;
  }
 
}  // end of loop

void recvBytesWithStartEndMarkers() {
    static boolean recvInProgress = false;
    static byte ndx = 0;
    byte startMarker = 0x3C;
    byte endMarker = 0x3E;
    byte rb;
   

    while (Serial.available() > 0 && newData == false) {
        setAmberLED();
        rb = Serial.read();

        if (recvInProgress == true) {
            #ifdef DEBUG 
              Serial.println("Receiving In Progress");
            #endif
            if (rb != endMarker) {
                receivedBytes[ndx] = rb;
                ndx++;
                if (ndx >= NUMBYTES) {
                    ndx = NUMBYTES - 1;
                }
            }
            else {
                #ifdef DEBUG              
                  Serial.println("End Marker");
                #endif
                //receivedBytes[ndx] = '\0'; // terminate the string
                recvInProgress = false;
                numReceived = ndx;  // save the number for use when printing
                ndx = 0;
                newData = true;
            }
        }

        else if (rb == startMarker) {
            #ifdef DEBUG
              Serial.println("Start Marker");
            #endif
            recvInProgress = true;
        }
    }
}

byte checkControl(byte ctrlBit){
  byte cmd=0x00;
  byte _ctrlBit=LOW;

  //Read control line state before writing to it.
  _ctrlBit=digitalRead(CTRL);

  //Setup control bit
  if(_ctrlBit==HIGH){
    cmd=0x80;
    #ifdef DEBUG
      Serial.println("This VB Control line high");
    #endif
  }else{
    cmd=0x00;
    #ifdef DEBUG
      Serial.println("This VB Control line low");
    #endif
  }
  
  if(ctrlBit==false){
    pinMode(CTRL,OUTPUT); //drive pin low
    #ifdef DEBUG
      Serial.println("Other VB Control line low");
    #endif
  }else{
    pinMode(CTRL,INPUT);  //hi-Z state
    #ifdef DEBUG
      Serial.println("Other VB Control line high");
    #endif
  }
  
  return cmd;
}

uint16_t calcCRC(byte* str, byte num)
{
  uint16_t crc=0; // starting value as you like, must be the same before each calculation
  for (int i=0;i<num;i++) // for each character in the string
  {
    crc= _crc16_update (crc, str[i]); // update the crc value
  }
  return crc;
}

/// --------------------------
/// Custom ISR Timer Routine
/// Triggers at the VBSyncClk rate
/// --------------------------
void timerIsr()
{
    // Toggle VBSyncClk
    digitalWrite( VBSyncClk, digitalRead( VBSyncClk ) ^ 1 );
    //Clear Amber or Green LED if its been set by receive or transmit on serial
    clearGreenLED();
    clearAmberLED();
}

void setupMasterSPI (void)
{
  SPCR=0x00;
  digitalWrite(SS, HIGH);  // ensure SS stays high for now
  // Put SCK, MOSI, SS pins into output mode
  // also put SCK, MOSI into LOW state, and SS into HIGH state.
  // Then put SPI hardware into Master mode and turn SPI on
  SPI.begin ();
  
}  // end of setup


void setupSlaveSPI (void)
{
  volatile char IOReg;
  // This tristates SS pin which if not done locks up Slave mode when coming
  //back from master mode.  
  digitalWrite(SS,LOW);
  pinMode(SS,INPUT);

  //Setup SPI control register
  SPCR |= (1<<CPOL); //Mode 3
  SPCR |= (1<<CPHA); //Mode 3
  SPCR &=~(1<<DORD); //MSB first
  SPCR |= (1<<SPE);  //enable SPI

  // have to send on master in, *slave out*
  pinMode(MISO, OUTPUT);
  
  SPCR |= (1<<SPIE); //enable SPI interrupt
  //SPI.attachInterrupt();
  IOReg   = SPSR; //required dummy read by datasheet
  IOReg   = SPDR; //required dummy read by datasheet

  // get ready for an interrupt 
  spi_slave_process_it = false;

}  // end of setup

// SPI slave receive interrupt routine
ISR (SPI_STC_vect)
{
  vbReturnByte = SPDR;  // grab byte from SPI Data Register
  spi_slave_process_it = true; //process this data
      
}  // end of interrupt routine SPI_STC_vect


//LED Helper Functions
void setMasterLED(void)
{
  digitalWrite( blueLED, LOW );
  digitalWrite( redLED, HIGH );
}

void setSlaveLED(void)
{
  digitalWrite( redLED, LOW );
  digitalWrite( blueLED, HIGH );
}

void clearRedBlueLED(void)
{
  digitalWrite( redLED, LOW );
  digitalWrite( blueLED, LOW );
}

void setGreenLED(void)
{
  digitalWrite( greenLED, HIGH );
}

void clearGreenLED(void)
{
  digitalWrite( greenLED, LOW );
}

void setAmberLED(void)
{
  digitalWrite( amberLED, HIGH );
}

void clearAmberLED(void)
{
  digitalWrite( amberLED, LOW );
}

void clearAmberGreenLED(void)
{
  digitalWrite( greenLED, LOW );
  digitalWrite( amberLED, LOW );
}


